
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#ifndef PG_APPLICATION_H
#define PG_APPLICATION_H

#include "xstd/NetAddr.h"
#include "xstd/String.h"
#include "xstd/Array.h"
#include "xstd/FileScanner.h"
#include "base/CmdLine.h"
#include "runtime/BcastRcver.h"

class OLog;
class Agent;
class AgentSym;
class AgentCfg;
class SizeOpt;
class NetAddrSym;

// holds information required during agent configuration process
class AgentCfg {
	public:
		AgentCfg();
		AgentCfg(AgentSym *aSym, const NetAddr &aHost);

	public:
		AgentSym *sym;
		NetAddr *host;
};


// common interface for polyclt and polysrv applications
class PolyApp: public FileScanTicker, public BcastRcver {
	public:
		typedef Array<const NetAddrSym*> AddrSyms;

	public:
		PolyApp();
		virtual ~PolyApp();

		virtual void noteMsgStrEvent(BcastChannel *ch, const char *msg);
		virtual void begCycle(int readyCount);
		virtual void endCycle();
		virtual bool tick();


		int run(int argc, char *argv[]);
		void flushState();

	protected:
		bool handleCmdLine(int argc, char *argv[]);
		void makeAgents();
		void reportRUsage() const;

		virtual void configure();
		virtual void reportCfg();
		virtual void logCfg();
		virtual void step();
		
		void configureBinLog(OLog &log, const String &fname, SizeOpt &bufSz);
		void configureLogs(int prec);
		void configureRnd();
		void configureHosts();
		void buildSchedule();
		void parseConfigFile(const String &fname);
		int scan(Time *toutp);
		
		void checkTiming(Time drift);

		void getIfaces();
		void getFakeIfaces();
		void getHostAddrs(Array<NetAddr*> &hosts) const;
		NetAddrSym *getAgentAddrMask() const;
		void makeAddresses();
		ostream &makeAddresses(Array<NetAddr*> &hosts, AddrSyms &agents, ostream &err);
		int makeAddresses(int hidx, AddrSyms &agents, int agentsPerHost);

		void getAgentAddrs(AddrSyms &agents) const;
		int countAddress(const NetAddr &addr) const;
		bool moreThanOneAddr() const;
		void addAgent(Agent *agent);
		
		virtual void startServices();
		virtual void startAgents();
		virtual void getOpts(Array<OptGrp*> &opts);
		virtual void wasIdle(Time tout);
		virtual void makeAgent(const AgentCfg &cfg) = 0;
		virtual void logState(OLog &log);

		virtual const String sideName() const = 0;
		virtual int logCat() const = 0;

	protected:
		CmdLine theCmdLine;

		String thePglCfg;            // temporary cache for logging purposes
		SchArray<in_addr> theIfaces; // list of interface or fake addresses
		Array<AgentCfg> theLocalCfgs;
		Array<Agent*> theLocals;
		String theAgentType;
		NetAddrSym *theAgentAddrMask;

		Time theIdleBeg;  // the start of an idle period
		Time theIdleEnd;  // beg + idle_tout
		bool isIdle;

		int theTickCount;  // see PolyApp::tick()
		int theStateCount; // number of states [flushed]
};

#endif
