
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "runtime/LogComment.h"
#include "pgl/RobotSym.h"
#include "runtime/StatPhase.h"
#include "runtime/StatPhaseMgr.h"
#include "client/AsyncClt.h"

AsyncClt::AsyncClt(RndDistr *aThinkDistr) {
	Assert(!theThinkDistr);
	Assert(aThinkDistr);
	theThinkDistr = aThinkDistr;
}

void AsyncClt::configure(const RobotSym *cfg, const NetAddr &aHost) {
	Assert(theThinkDistr);
	Client::configure(cfg, aHost);
}

void AsyncClt::wakeUp(const Alarm &a) {
	Client::wakeUp(a);
	scheduleLaunch(a.time()); // next

	if (theExtraLaunchLvl)
		--theExtraLaunchLvl;
	else
		tryLaunch(genXact());
}

void AsyncClt::scheduleLaunch(Time lastLaunch) {
	Assert(!thePendAlarmCnt);
	Assert(theThinkDistr);

	const double lf = TheStatPhaseMgr->loadFactor().current(lastLaunch, theLaunchCount);
	Assert(lf >= 0);

	if (lf <= 0) // zero load: no more requests needed
		return;

	double delay;
	if (TheStatPhaseMgr->loadFactor().flat()) {
		delay = theThinkDistr->trial()/lf;
	} else {
		// next launch should happen [0, 2dt] from last launch
		// where dt is 1/current_rate
		static RndGen rng;
		delay = rng(0.0, 2.*theThinkDistr->mean()/lf);
	}
	sleepTill(lastLaunch + Time::Secd(delay));
}

void AsyncClt::launchFailed(CltXact *x) {
	Client::launchFailed(x);

	// if no alarms are pending, we may get stuck and submit
	// no more xactions (but this should not happen with AsyncClt)
	Should(thePendAlarmCnt || theCcXactLvl);
}
