
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#ifndef PG_CLIENT_CLASS_H
#define PG_CLIENT_CLASS_H

#include "xstd/AlarmClock.h"
#include "xstd/Ring.h"
#include "runtime/XactFarm.h"
#include "runtime/Agent.h"
#include "runtime/BcastRcver.h"
#include "client/IcpCltXact.h"
#include "client/PrivWorld.h"

class Connection;
class CltConnMgr;
class DnsMgr;
class CltCfg;
class CltSharedCfgs;
class PortMgr;
class ServerRep;
class ObjId;
class ObjId;
class PrivCache;
class DistrPoint;
class RobotSym;

class Client: public Agent, public AlarmUser, public BcastRcver {
	public:
		static void Farm(XactFarm<CltXact> *aFarm);
		static void LogState(OLog &log);
		static ServerRep *GetServerRep(int idx);

	public:
		Client();
		virtual ~Client();

		virtual void configure(const RobotSym *cfg, const NetAddr &aHost);
		void icpClient(IcpClient *anIcpClient);

		virtual void start();

		// called by the Cache on a miss
		CltXact *fetch(const ObjId &oid, DistrPoint *dp);

		virtual void noteEmbedded(CltXact *x, const ObjId &embedOid);
		virtual void noteRedirect(CltXact *x, const ObjId &destOid);
		virtual void noteXactDone(CltXact *x);
		virtual void notePeerAsked(IcpCltXact *q);
		virtual void noteAddrLookup(const NetAddr &addr, CltXact *x);

		void selectTarget(ObjId &oid); // embedded objs and redirects need it

		IcpClient *icpClient() { return theIcpClient; }
		PrivCache *privCache() { return thePrivCache; }
		const NetAddr proxy() const { return theProxyAddr; }

		virtual int logCat() const;

	protected:
		static void FreezeSrvWss();

		PortMgr *getPortMgr(bool bind2iface);

		bool tryLaunch(CltXact *x);
		void suspendXact(CltXact *x);
		void resumeXact();

		CltXact *genXact();
		CltXact *genXact(const ObjId &oid);
		void genOid(ObjId &oid);
		void selectViserv(ObjId &oid);
		int selectAnyTarget(int visName);
		int selectTypedTarget(ObjId &oid, int visName);
		void selectObj(ObjId &oid);
		void selectContType(ObjId &oid);
		void selectReqType(ObjId &oid);

		const NetAddr &nextHopAddr(const CltXact *x) const;
		void askPeer(const NetAddr &addr, CltXact *x);
		void lookupAddr(CltXact *x);

		virtual void scheduleLaunch(Time lastLaunch) = 0;
		virtual void launchFailed(CltXact *x) = 0; // with body

		virtual void noteInfoEvent(BcastChannel *ch, InfoEvent ev);
		virtual void noteLogEvent(BcastChannel *ch, OLog &log);

		virtual void describe(ostream &os) const;

	private:
		bool launch(CltXact *x);

	protected:
		static XactFarm<CltXact> *TheXacts;
		static ObjFarm<IcpCltXact> TheIcpXacts;

		static CltSharedCfgs *TheSharedCfgs;
		static Array<PortMgr*> ThePortMgrs;  // port managers for all aliases
		static int TheClientCnt;             // global client counter

		Array<PrivWorld> thePrivWorlds; // per-origin private worlds
		PrivCache *thePrivCache;        // private cache for this agent

		Array<CltXact*> theLaunchDebts; // postponed requests

		CltCfg *theCfg;         // client configuration (maybe shared)
		CltConnMgr *theConnMgr; // one for all servers
		DnsMgr *theDnsMgr;      // DNS client, cache, etc.

		NetAddr theProxyAddr;
		IcpClient *theIcpClient;

		int theCcXactLvl;      // outstanding concurrent xactions level
		int theLaunchCount;    // total number of xaction launches
		int theExtraLaunchLvl; // uncompensated out-of-order launches
};

#endif
