
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "xstd/gadgets.h"
#include "xstd/rndDistrs.h"
#include "base/OLog.h"
#include "base/polyLogCats.h"
#include "runtime/LogComment.h"
#include "runtime/SharedOpts.h"
#include "runtime/StatIntvl.h"
#include "runtime/Rptmstat.h"
#include "runtime/globals.h"
#include "pgl/PglSemx.h"
#include "pgl/RobotSym.h"
#include "pgl/BenchSym.h"
#include "pgl/BenchSideSym.h"
#include "client/AsyncClt.h"
#include "client/SyncClt.h"
#include "client/PassClt.h"
#include "client/ServerRep.h"
#include "client/IcpCltXact.h"
#include "client/CltOpts.h"
#include "client/CltXact.h"
#include "client/CltConnMgr.h"
#include "client/wssFreezers.h"
#include "client/PolyClt.h"

typedef XactFarmT<CltXact, CltXact> CltXactFarm;


PolyClt::PolyClt() {
	theAgentType = "Robot";
}

void PolyClt::makeAgent(const AgentCfg &cfg) {
	Assert(cfg.sym);
	
	const RobotSym *rcfg = &(const RobotSym&)cfg.sym->cast("Robot");

	Client *client = 0;
	RndDistr *iad = 0;
	if (rcfg->reqInterArrival(iad)) {
		if (iad)
			client = new AsyncClt(iad);
		else
			client = new PassClt();
	} else {
		client = new SyncClt();
	}

	Assert(client);
	client->configure(rcfg, *cfg.host);

	addAgent(client);
}

void PolyClt::configure() {

	PolyApp::configure();

	TheWssFreezer = 0;
	if (PglSemx::TheWorkSetLen >= 0 && PglSemx::TheWorkSetCap >= 0) {
		cerr << ThePrgName << ": canot handle more than one WSS freezing conditions" << endl;
		exit(-3);
	}
	if (PglSemx::TheWorkSetLen >= 0)
		TheWssFreezer = new TimeWssFreezer(PglSemx::TheWorkSetLen);
	if (PglSemx::TheWorkSetCap >= 0)
		TheWssFreezer = new FillWssFreezer(PglSemx::TheWorkSetCap);

	Rptmstat::IsEnabled = true; // only clinet-side cares about load

	Client::Farm(new CltXactFarm);
	IcpCltXact::TheTimeout = TheCltOpts.theIcpTout;
	StatIntvl::ActiveCat(lgcCltSide);
	StatIntvl::TheReportCat = lgcCltSide;
}

void PolyClt::getOpts(Array<OptGrp*> &opts) {
	PolyApp::getOpts(opts);
	opts.append(&TheCltOpts);
}

void PolyClt::reportCfg() {
	PolyApp::reportCfg();
}

void PolyClt::logState(OLog &log) {
	PolyApp::logState(log);
	Client::LogState(log);
}

void PolyClt::startServices() {
	PolyApp::startServices();
	if (TheWssFreezer)
		TheWssFreezer->start();
}

const String PolyClt::sideName() const {
	return "client";
}

int PolyClt::logCat() const {
	return lgcCltSide;
}

int main(int argc, char *argv[]) {
	PolyClt app;
	return app.run(argc, argv);
}
