
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */


#include "base/polygraph.h"

#include "xstd/h/iostream.h"
#include "xstd/h/strstream.h"

#include "base/BStream.h"
#include "runtime/IOBuf.h"
#include "csm/ContentDbase.h"

#include "csm/cdbEntries.h"
#include "xstd/gadgets.h"


/* CdbEntryPrnOpt */

CdbEntryPrnOpt::CdbEntryPrnOpt(): buf(0), injector(0) {
}

/* CdbEntry */

void CdbEntry::print(CdbEntryPrnOpt &opt) const {
	ostrstream os(opt.buf->space(), Min(opt.sizeMax, opt.buf->spaceSize()));
	print(os);
	opt.buf->appended(os.tellp());
}


/* ContentDbase */

ContentDbase::ContentDbase() {
}

ContentDbase::~ContentDbase() {
	while (theEntries.count()) delete theEntries.pop();
}

int ContentDbase::count() const {
	return theEntries.count();
}

static
IBStream &operator >>(IBStream &is, CdbEntry *&e) {
	const int type = is.geti();
	if (is) {
		e = ContentDbase::CreateEntry(type);
		e->load(is);
	} else {
		e = 0;
	}
	return is;
}

static
OBStream &operator <<(OBStream &os, const CdbEntry *e) {
	Assert(e);
	os << e->type();
	e->store(os);
	return os;
}

void ContentDbase::load(IBStream &is) {
	theName = is.name();
	is >> theEntries;
}

void ContentDbase::store(OBStream &os) const {
	theName = os.name();
	os << theEntries;
}

void ContentDbase::add(CdbEntry *e) {
	theEntries.append(e);
}

ostream &ContentDbase::print(ostream &os) const {
	for (int i = 0; i < count(); ++i)
		theEntries[i]->print(os);
	return os;
}

CdbEntry *ContentDbase::CreateEntry(int type) {
	switch (type) {
		case cdbeText:
			return new CdbeText;
		case cdbeTag:
			return new CdbeTag;
		default:
			Assert(false); // unknown entry type
	}
	return 0;
}
