
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "xstd/String.h"
#include "xstd/TblDistr.h"
#include "pgl/PglRec.h"
#include "pgl/PglDistrSym.h"
#include "pgl/PglStringSym.h"
#include "pgl/PglArraySym.h"
#include "pgl/SocketSym.h"
#include "pgl/UniqIdSym.h"
#include "pgl/PopModelSym.h"
#include "pgl/AgentSym.h"


AgentSym::AgentSym(const String &aType): HostsBasedSym(aType) {
	theRec->bAdd("Socket", "socket", new SocketSym);

	theRec->bAdd("string", "kind", 0);
	theRec->bAdd("uniq_id", "world", 0);
	theRec->bAdd("time_distr", "xact_think", 0);
	theRec->bAdd("int_distr", "pconn_use_lmt", 0);
	theRec->bAdd("time", "idle_pconn_tout", 0);
	theRec->bAdd("PopModel", "pop_model", new PopModelSym);
}

AgentSym::AgentSym(const String &aType, PglRec *aRec): HostsBasedSym(aType, aRec) {
}

bool AgentSym::isA(const String &type) const {
	return HostsBasedSym::isA(type) || type == "Agent";
}

SynSym *AgentSym::dupe(const String &type) const {
	if (isA(type))
		return new AgentSym(this->type(), theRec->clone());
	return HostsBasedSym::dupe(type);
}

String AgentSym::kind() const {
	return getString("kind");
}

UniqId AgentSym::world() const {
	SynSymTblItem *wi = 0;
	Assert(theRec->find("world", wi));
	return wi->sym() ?
		((const UniqIdSym&)wi->sym()->cast("uniq_id")).val() : UniqId();
}

SocketSym *AgentSym::socket() const {
	SynSymTblItem *si = 0;
	Assert(theRec->find("socket", si));
	Assert(si->sym());
	return &(SocketSym&)si->sym()->cast("Socket");
}

RndDistr *AgentSym::xactThink() const {
	return getDistr("xact_think");
}

RndDistr *AgentSym::pconnUseLmt() const {
	return getDistr("pconn_use_lmt");
}

Time AgentSym::idlePconnTimeout() const {
	return getTime("idle_pconn_tout");
}

PopModelSym *AgentSym::popModel() const {
	SynSymTblItem *pmi = 0;
	Assert(theRec->find("pop_model", pmi));
	if (!pmi->sym())
		return 0;

	PopModelSym *pms = &(PopModelSym&)pmi->sym()->cast("PopModel");
	if (pms->empty())
		return 0;
	return pms;
}

RndDistr *AgentSym::namesToDistr(const String &fieldName, const Array<String*> &names) const {
	Assert(names.count());

	SynSymTblItem *fti = 0;
	Assert(theRec->find(fieldName, fti));
	if (!fti->sym())
		return 0; // undefined

	// build selector using label index
	ArraySym &nts = (ArraySym&)fti->sym()->cast("Array");
	Array<double> probs_in(names.count());
	Array<double> probs_out(names.count());
	nts.copyProbs(probs_in);

	for (int i = 0; i < probs_in.count(); ++i) {
		const String &name = ((const StringSym&)nts[i]->cast("string")).val();

		int idx = -1;
		for (int t = 0; idx < 0 && t < names.count(); ++t) {
			if (*names[t] == name)
				idx = t;
		}
			
		if (idx < 0) {
			cerr << loc() << "unknown quoted name `" << name
				<< "'; known names within " << fieldName << " context are:";
			for (int n = 0; n < names.count(); ++n)
				cerr << " \"" << *names[n] << '"';
			cerr << endl;
			exit(-2);
		}

		if (probs_in[i] > 0)
			probs_out.put(probs_in[i], idx);
	}

	if (!probs_out.count())
		return 0;

	return TblDistr::FromDistrTable(kind() + "-" + fieldName, probs_out);
}
