
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "xstd/h/strstream.h"
#include "xstd/h/math.h"

#include "xstd/TblDistr.h"
#include "pgl/PglIntSym.h"
#include "pgl/PglNetAddrRange.h"
#include "pgl/PglNetAddrSym.h"
#include "pgl/PglNetAddrArraySym.h"

#include "xstd/rndDistrs.h"
#include "xstd/gadgets.h"


NetAddrArraySym::NetAddrArraySym(): ArraySym("addr") {
}

ArraySym *NetAddrArraySym::cloneType(const String &) const {
	return new NetAddrArraySym();
}

ostream &NetAddrArraySym::print(ostream &os, const String &pfx) const {
	if (probsVary())
		return ArraySym::print(os, pfx);

	Array<PglNetAddrRange*> sums;
	sums.append(new PglNetAddrRange);


	// first scan: NetAddrSyms -> fewer PglNetAddrRanges
	for (int i = 0; i < count(); ++i) {
		Assert(item(i));
		Assert(item(i)->isA("addr"));
		const NetAddrSym &nas = (const NetAddrSym&)item(i)->cast("addr");
		PglNetAddrRange r;
		addrSymToRange(nas, r);
		if (!sums.last()->canMerge(r))
			sums.append(new PglNetAddrRange);
		sums.last()->merge(r);
	}

	// secondary scans: PglNetAddrRanges -> fewer PglNetAddrRanges
	bool merged;
	do {
		merged = false;
		for (int x = 0, y = 1; !merged && y < sums.count(); ++y) {
			PglNetAddrRange *&r = sums[y];
			if (!r)
				continue;
			if (sums[x] && sums[x]->canMerge(*r)) {
				sums[x]->merge(*r);
				delete r;
				r = 0;
				merged = true;
			} else {
				++x;
			}
		}
	} while (merged);

	// print and destroy
	os << '[';
	int prnCount = 0;
	for (int k = 0; k < sums.count(); ++k) {
		if (PglNetAddrRange *r = sums[k]) {
			printRange(os, prnCount++, *r, -1);
			delete r;
		}
	}
	return os << ']';
}

void NetAddrArraySym::printRange(ostream &os, int idx, const PglNetAddrRange &r, double prob) const {
	if (idx)
		os << ',';
	os << '\'' << r.toStr() << '\'';
	if (prob >= 0)
		os << ':' << (prob*100) << '%';
}

void NetAddrArraySym::addrSymToRange(const NetAddrSym &nas, PglNetAddrRange &r) const {
	ostrstream buf;
	nas.printUnquoted(buf);
	buf << ends;
	Assert(r.parse(buf.str()));
	ostrstreamFreeze(buf, true);
}
