
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "pgl/PglNumSym.h"
#include "pgl/PglTimeSym.h"
#include "pgl/PglSizeSym.h"
#include "pgl/PglBwidthSym.h"
#include "pgl/PglRateSym.h"

#include "xstd/gadgets.h"

RateSym::RateSym(double aCount, Time anInterval): ExpressionSym("rate"),
	theCount(aCount), theInterval(anInterval) {
}

RateSym::RateSym(double aVal): ExpressionSym("rate") {
	val(aVal);
}

bool RateSym::isA(const String &type) const {
	return ExpressionSym::isA(type) ||
		type == "Rate" || type == "rate";
}

SynSym *RateSym::dupe(const String &type) const {
	if (isA(type))
		return new RateSym(theCount, theInterval);
	return ExpressionSym::dupe(type);
}

double RateSym::val() const {
	return Ratio(theCount, theInterval.secd());
}

void RateSym::val(double aVal) {
	theCount = aVal;
	theInterval = Time::Sec(1);
}

ExpressionSym *RateSym::unOper(const String &op) const {
	if (*op == '+')
		return new RateSym(+theCount, theInterval);
	else
	if (*op == '-')
		return new RateSym(-theCount, theInterval);
	else
		return ExpressionSym::unOper(op);
}

ExpressionSym *RateSym::bnOper(const String &op, const SynSym &s) const {
	if (op.len() != 1)
		return ExpressionSym::bnOper(op, s);

	if (RateSym *rs = (RateSym*)s.clone("rate"))
		return operRR(op, rs, s);

	if (TimeSym *ts = (TimeSym*)s.clone("time"))
		return operRT(op, ts, s);

	if (SizeSym *ss = (SizeSym*)s.clone("size"))
		return operRS(op, ss, s);

	if (NumSym *ns = (NumSym*)s.clone("float"))
		return operRN(op, ns, s);

	return ExpressionSym::bnOper(op, s);
}

// rate op rate
ExpressionSym *RateSym::operRR(const String &op, RateSym *rs, const SynSym &s) const {
	double otherVal = rs->val();
	delete rs;

	if (*op == '/')
		checkDenom(otherVal);

	switch (*op) {
		case '+': return new RateSym(val() + otherVal);
		case '-': return new RateSym(val() - otherVal);
		case '/': return new NumSym(val() / otherVal);
		default:  return ExpressionSym::bnOper(op, s);
	}
}

ExpressionSym *RateSym::operRS(const String &op, SizeSym *szs, const SynSym &s) const {
	BigSize sz = szs->val();
	delete szs;

	switch (*op) {
		case '*': return new BwidthSym(Bandwidth(sz*theCount, theInterval));
		default: return ExpressionSym::bnOper(op, s);
	}
}

ExpressionSym *RateSym::operRT(const String &op, TimeSym *ts, const SynSym &s) const {
	Time otherInterval = ts->val();
	delete ts;

	switch (*op) {
		case '*': return new NumSym(theCount*(otherInterval/theInterval));
		default: return ExpressionSym::bnOper(op, s);
	}
}

ExpressionSym *RateSym::operRN(const String &op, NumSym *ns, const SynSym &s) const {
	double otherCount = ns->val();
	delete ns;
	
	if (*op == '/')
		checkDenom(otherCount);

	switch (*op) {
		case '*': return new RateSym(theCount*otherCount, theInterval);
		case '/': return new RateSym(theCount/otherCount, theInterval);
		default: return ExpressionSym::bnOper(op, s);
	}
}

ostream &RateSym::print(ostream &os, const String &) const {
	return os << val() << "/sec";
}
