
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "xstd/h/math.h"

#include "pgl/PglNumSym.h"
#include "pgl/PglTimeSym.h"


TimeSym::TimeSym(Time aVal): ExpressionSym("time"), theVal(aVal) {
}

bool TimeSym::isA(const String &type) const {
	return ExpressionSym::isA(type) ||
		type == "Time" || type == "time";
}


SynSym *TimeSym::dupe(const String &type) const {
	if (isA(type))
		return new TimeSym(theVal);
	return ExpressionSym::dupe(type);
}

ExpressionSym *TimeSym::unOper(const String &op) const {
	if (*op == '+')
		return new TimeSym(+theVal);
	else
	if (*op == '-')
		return new TimeSym(-theVal);
	else
		return ExpressionSym::unOper(op);
}


ExpressionSym *TimeSym::bnOper(const String &op, const SynSym &s) const {
	if (op.len() != 1)
		return ExpressionSym::bnOper(op, s);

	// division is defined for both Time and Num
	const bool forceTT = *op == '/' && s.canBe("time");

	if (*op == '+' || *op == '-' || forceTT)
		return operTT(op, s); // time op time

	if (*op == '*' || *op == '/')
		return operTN(op, s); // time op num
	
	return ExpressionSym::bnOper(op, s);
}

ExpressionSym *TimeSym::operTT(const String &op, const SynSym &s) const {
	Time otherVal;
	
	if (s.isA("time")) {
		otherVal = ((const TimeSym&)s.cast("time")).val();
	} else {
		TimeSym *ts = (TimeSym*)s.clone("time");
		if (!ts)
			return ExpressionSym::bnOper(op, s);
		otherVal = ts->val();
		delete ts;
	}

	if (*op == '/' && otherVal == 0)
		cerr << theLoc << "denominator is zero" << endl << xexit;

	switch (*op) {
		case '+': return new TimeSym(theVal + otherVal);
		case '-': return new TimeSym(theVal - otherVal);
		//case '*': return new TimeSym(theVal * otherVal);
		case '/': return new NumSym(theVal / otherVal);
		default: return ExpressionSym::bnOper(op, s);
	}
}

ExpressionSym *TimeSym::operTN(const String &op, const SynSym &s) const {
	double otherVal = 0;
	
	if (s.isA("float")) {
		otherVal = ((const NumSym&)s.cast("float")).val();
	} else {
		NumSym *fs = (NumSym*)s.clone("float");
		if (!fs)
			return ExpressionSym::bnOper(op, s);
		otherVal = fs->val();
		delete fs;
	}

	if (*op == '/' && fabs(otherVal) < 1e-8)
		cerr << theLoc << "denominator is "
			<< (otherVal ? "[virtually]" : "")
			<< " zero" << endl << xexit;

	switch (*op) {
		//case '+': return new TimeSym(theVal + otherVal);
		//case '-': return new TimeSym(theVal - otherVal);
		case '*': return new TimeSym(theVal * otherVal);
		case '/': return new TimeSym(theVal / otherVal);
		default: return ExpressionSym::bnOper(op, s);
	}
}

ostream &TimeSym::print(ostream &os, const String &) const {
	return os << theVal;
}
