
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "xstd/String.h"
#include "xstd/rndDistrs.h"
#include "pgl/PglBoolSym.h"
#include "pgl/PglIntSym.h"
#include "pgl/PglNumSym.h"
#include "pgl/PglRateSym.h"
#include "pgl/PglRec.h"
#include "pgl/PglStringSym.h"
#include "pgl/PglNetAddrSym.h"
#include "pgl/PglArraySym.h"
#include "pgl/DnsResolverSym.h"
#include "pgl/RobotSym.h"


RobotSym::RobotSym(): AgentSym("Robot") {
	theRec->bAdd("addr[]", "origins", 0);
	theRec->bAdd("time", "launch_win", 0);
	theRec->bAdd("rate", "req_rate", 0);
	theRec->bAdd("time_distr", "req_inter_arrival", 0);
	theRec->bAdd("float", "public_interest", 0);
	theRec->bAdd("float", "recurrence", 0);
	theRec->bAdd("float", "embed_recur", 0);
	theRec->bAdd("string[]", "req_types", 0);
	theRec->bAdd("int", "private_cache_cap", 0);
	theRec->bAdd("bool", "unique_urls", 0);
	theRec->bAdd("int", "open_conn_lmt", 0);
	theRec->bAdd("int", "wait_xact_lmt", 0);
	theRec->bAdd("float", "minimize_new_conn", 0);
	theRec->bAdd("int", "icp_port", 0);
	theRec->bAdd("addr", "peer_http", 0);
	theRec->bAdd("addr", "peer_icp", 0);
	theRec->bAdd("DnsResolver", "dns_resolver", new DnsResolverSym);
}

RobotSym::RobotSym(const String &aType, PglRec *aRec): AgentSym(aType, aRec) {
}

bool RobotSym::isA(const String &type) const {
	return AgentSym::isA(type) || type == "Robot";
}

SynSym *RobotSym::dupe(const String &type) const {
	if (isA(type))
		return new RobotSym(this->type(), theRec->clone());
	return AgentSym::dupe(type);
}

bool RobotSym::reqRate(double &rate) const {
	return getRate("req_rate", rate);
}

bool RobotSym::reqInterArrival(RndDistr *&iad) const {
	iad = getDistr("req_inter_arrival");
	double rr = -1;
	const bool hasRR = reqRate(rr);
	if (hasRR && iad)
		cerr << loc() << ": cannot specify both request rate"
			<< " and inter-arrival distribution" << endl << xexit;

	if (iad)
		return true;

	if (!hasRR)
		return false;

	if (rr <= 0)
		return true; // iad is nil

	// poisson request stream (default) is modeled using an exponential
	// distribution with a mean of a given request rate
	static RndGen rng; // all clients should share one rng
	iad = new ExpDistr(&rng, 1/rr);
	return true;
}

Time RobotSym::launchWin() const {
	return getTime("launch_win");
}

bool RobotSym::publicRatio(double &ratio) const {
	return getDouble("public_interest", ratio);
}

bool RobotSym::recurRatio(double &ratio) const {
	return getDouble("recurrence", ratio);
}

bool RobotSym::embedRecurRatio(double &ratio) const {
	return getDouble("embed_recur", ratio);
}

bool RobotSym::uniqueUrls(bool &set) const {
	return getBool("unique_urls", set);
}

bool RobotSym::openConnLimit(int &lmt) const {
	return getInt("open_conn_lmt", lmt);
}

bool RobotSym::waitXactLimit(int &lmt) const {
	return getInt("wait_xact_lmt", lmt);
}

bool RobotSym::minimizeNewConn(double &prob) const {
	return getDouble("minimize_new_conn", prob);
}

bool RobotSym::icpPort(int &port) const {
	return getInt("icp_port", port);
}

NetAddr RobotSym::peerHttp() const {
	return getNetAddr("peer_http");
}

NetAddr RobotSym::peerIcp() const {
	return getNetAddr("peer_icp");
}

DnsResolverSym *RobotSym::dnsResolver() const {
	SynSymTblItem *di = 0;
	Assert(theRec->find("dns_resolver", di));
	return di->sym() ?
		&((DnsResolverSym&)di->sym()->cast("DnsResolver")) : 0;
}

bool RobotSym::privCache(int &capacity) const {
	return getInt("private_cache_cap", capacity);
}

bool RobotSym::proxies(Array<NetAddr*> &) const {
	return false;
}

bool RobotSym::origins(Array<NetAddr*> &addrs) const {
	return getNetAddrs("origins", addrs);
}

bool RobotSym::origins(Array<NetAddr*> &addrs, RndDistr *&selector) const {
	SynSymTblItem *oi = 0;
	Assert(theRec->find("origins", oi));
	if (!oi->sym())
		return false; // undefined

	// build address array and selector
	origins(addrs);
	ArraySym &os = (ArraySym&)oi->sym()->cast("Array");
	selector = os.makeSelector(kind() + "-origins");
	return true;
}

RndDistr *RobotSym::reqTypes(const Array<String*> &typeNames) const {
	return namesToDistr("req_types", typeNames);
}
