// (C) 2001 The Measurement Factory

#if 0
g++ -O3 -g -DHAVE_CONFIG_H -I.. -I../.. -Wall -o connprobe connprobe.cc \
	../runtime/libruntime.a ../pgl/libpgl.a ../base/libbase.a \
	../xparser/libxparser.a ../xstd/libxstd.a
#endif


#define _IP_VHL 1

#include "base/polygraph.h"

#include <stdlib.h>
#include <fstream.h>

#include "xstd/h/sys/types.h"
#include "xstd/h/os_std.h"
#include "xstd/h/sys/socket.h"
#include "xstd/h/netinet.h"

#include <sys/param.h>
#include <sys/mbuf.h>

#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <net/route.h>
#include <netinet/ip_fw.h>
#include <netinet/ip_dummynet.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>

#include "xstd/h/string.h"
#include "xstd/h/fcntl.h"
#include "xstd/h/iomanip.h"

#include "xstd/IpSocket.h"
#include "xstd/Clock.h"
#include "xstd/Error.h"
#include "xstd/Rnd.h"
#include "xstd/gadgets.h"

#include "base/CmdLine.h"
#include "base/opts.h"
#include "base/polyOpts.h"
#include "base/StatIntvlRec.h"
#include "base/polyLogCats.h"
#include "pgl/PglNetAddrRange.h"
#include "runtime/Notifier.h"

struct pseudo_hdr {
	struct in_addr src;
	struct in_addr dst;
	unsigned char zero;
	unsigned char proto;
	unsigned short len;
};

struct ConnState {
	tcp_seq seq;
	tcp_seq ack;
	char tcp_state;
};


class IntRange {
	public:
		IntRange(int a = 0, int b = 0): min(a), max(b) {}

		int	count() const { return max - min + 1; }

	public:
		int min;
		int max;
};

class AddrHashItem {
	public:
		typedef int Data;

	public:
		AddrHashItem(): theAddr(0) {}
		AddrHashItem(const NetAddr &a, const Data &d): theAddr(0) { setAddr(a, d); }

		void destroy() { delete theAddr; theAddr = 0; }

		const in_addr &addrN() const { return theAddr->addrN(); }
		bool hasAddr(const NetAddr &a, Data &d) const;
		int portCount() const { return thePorts.count(); }

		NetAddr addrAt(int idx) const { return NetAddr(addrN(), portAt(idx)); }
		Data dataAt(int idx) const { return theDatas.min + idx; }

		void addAddr(const NetAddr &a, const Data &d);

	protected:
		void setAddr(const NetAddr &a, const Data &d);
		int portAt(int idx) const { return thePorts.min + idx; }

	protected:
		NetAddr *theAddr;
		IntRange thePorts;
		IntRange theDatas;
};

class AddrHash;

class AddrHashIter {
	public:
		typedef AddrHashItem Item;
		typedef AddrHashItem::Data Data;

	public:
		AddrHashIter(AddrHash &aHash);

		operator void *() const { return atEnd() ? 0 : ((void*)-1); }
		bool atEnd() const;

		NetAddr addr() const;
		Data data() const;

		AddrHashIter &operator ++() { ++thePort; sync(); return *this; }

		void rewind();

	protected:
		void sync();

	protected:
		AddrHash &theHash;
		int theLine;
		int theOff;
		int thePort;
};

class AddrHash {
	public:
		typedef AddrHashItem Item;
		typedef AddrHashItem::Data Data;
		friend class AddrHashIter;

	public:
		AddrHash();
		~AddrHash();

		void capacity(int aCap);
		int count() const { return theCount; }
		int maxChainLen() const { return theMaxChainLen; }
		double utilization() const;

		void add(const NetAddr &addr, const Data &data);
		bool find(const NetAddr &addr, Data &data) const;

		AddrHashIter iterator();

	protected:
		typedef Array<Item> Line;

	protected:
		int hash(const NetAddr &addr) const;
		AddrHashItem *getItem(Line *line, const NetAddr &addr);

	protected:
		Array<Line*> theIndex;
		int theCount;        // total number of entries
		int theMaxChainLen;
};


class ConnProbeOpts: public OptGrp {
	public:
		ConnProbeOpts();

		virtual bool validate() const;

		virtual ostream &printAnonym(ostream &os) const;
		virtual bool parseAnonym(const Array<const char *> &opts);
		virtual bool canParseAnonym() const { return true; }

	public:
		HelpOpt theHelpOpt;
		StrOpt theSideName;
		StrArrOpt theClients;
		StrArrOpt theServers;
		RateOpt thePeakRate;
		TimeOpt theStuckTout;
		NetAddrOpt theNotifAddr;
		StrOpt theConsFileName;
};

#define IPHDRLEN 20
#define TCPHDRLEN 20
#define IPPROTO_THIS 252

static char pktbuf[1500];
static tcp_seq myseq;
static IpSocket raw;
static Array<ConnState> theConnections;
static AddrHash theClients;
static Array<NetAddr*> theServers;
static ConnProbeOpts TheConnProbeOpts;
static CmdLine theCmdLine;
static String thePrgName;
static bool isClient = false;
static int nestablished = 0;
static int npktssent = 0;
static int npktsrecv = 0;
static Time theLastEstbTime;
static StatIntvlRec TheStats;
static Notifier *TheNotifier = 0;


/* AddrHashItem */

bool AddrHashItem::hasAddr(const NetAddr &a, Data &d) const {
	Assert(addrN() == a.addrN());
	if (thePorts.min <= a.port() && a.port() <= thePorts.max) {
		d = theDatas.min + (a.port() - thePorts.min);
		return true;
	}
	return false;
}

void AddrHashItem::addAddr(const NetAddr &a, const Data &d) {
	Assert(addrN() == a.addrN());
	Assert(thePorts.max+1 == a.port());
	Assert(theDatas.max+1 == d);
	thePorts.max++;
	theDatas.max++;
}

void AddrHashItem::setAddr(const NetAddr &a, const Data &d) {
	Assert(!theAddr);
	theAddr = new NetAddr(a);
	thePorts.min = thePorts.max = a.port();
	theDatas.min = theDatas.max = d;
}


/* AddrHashIter */

AddrHashIter::AddrHashIter(AddrHash &aHash): theHash(aHash), theLine(0), theOff(0), thePort(0) {
	sync();
}

bool AddrHashIter::atEnd() const {
	return theLine >= theHash.theIndex.capacity();
}

NetAddr AddrHashIter::addr() const {
	return theHash.theIndex[theLine]->item(theOff).addrAt(thePort);
}
 
AddrHashIter::Data AddrHashIter::data() const {
	return theHash.theIndex[theLine]->item(theOff).dataAt(thePort);
}

void AddrHashIter::sync() {
	while (theLine < theHash.theIndex.capacity()) {
		if (AddrHash::Line *line = theHash.theIndex[theLine]) {
			if (theOff < line->count()) {
				const Item &i = theHash.theIndex[theLine]->item(theOff);
				if (thePort < i.portCount())
					break;
				theOff++;
				thePort = 0;
			}
		}
		theLine++;
		theOff = 0;
		thePort = 0;
	}
}

void AddrHashIter::rewind() {
	theLine = theOff = thePort = 0;
	sync();
}


/* AddrHash */

AddrHash::AddrHash(): theCount(0), theMaxChainLen(0) {
}

AddrHash::~AddrHash() {
	while (theIndex.count()) {
		Line *line = theIndex.pop();
		while (line && line->count())
			line->pop().destroy();
		delete line;
	}
}

void AddrHash::capacity(int aCap) {
	Assert(aCap >= theIndex.capacity());
	theIndex.stretch(aCap | 1);
	theIndex.count(theIndex.capacity());
}

double AddrHash::utilization() const {
	int slotCount = 0;
	int addrCount = 0;
	for (int i = 0; i < theIndex.count(); ++i) {
		if (Line *line = theIndex[i]) {
			slotCount += line->count();
			addrCount += line->count();
		} else {
			slotCount++;
		}
	}
	return Percent(addrCount, slotCount);
}

void AddrHash::add(const NetAddr &addr, const Data &data) {
	Assert(theIndex.capacity());

	Data dummy;
	if (find(addr, dummy))
		cerr << here << "duplicate address: " << addr << endl << xexit;

	const int lidx = hash(addr);
	if (Line *line = theIndex[lidx]) {
		if (Item *i = getItem(line, addr))
			i->addAddr(addr, data);
		else
			line->append(Item(addr, data));
	} else {
		theIndex[lidx] = new Line(4);
		theIndex[lidx]->append(Item(addr, data));
	}
	theCount++;
	theMaxChainLen = Max(theMaxChainLen, theIndex[lidx]->count());
}

AddrHashItem *AddrHash::getItem(Line *line, const NetAddr &addr) {
	for (int i = 0; i < line->count(); ++i) {
		if (line->item(i).addrN() == addr.addrN())
			return &line->item(i);
	}
	return 0;
}

bool AddrHash::find(const NetAddr &addr, Data &data) const {
	if (!theIndex.capacity())
		return false;

	const int lidx = hash(addr);
	if (const Line *line = theIndex[lidx]) {
		for (int i = 0; i < line->count(); ++i) {
			if (line->item(i).hasAddr(addr, data))
				return true;
		}
	}
	return false;
}

AddrHashIter AddrHash::iterator() {
	return AddrHashIter(*this);
}

int AddrHash::hash(const NetAddr &addr) const {
	const int h = (int)(addr.lna() + addr.net());
	return (h < 0 ? h + INT_MAX : h) % theIndex.capacity();
}


/* ConnProbeOpts */

ConnProbeOpts::ConnProbeOpts():
	theHelpOpt(this,  "help",		 "list of options"),
	theSideName(this, "side <client|server>", "operation mode"),
	theClients(this, "clients <str>", "clients ip(s) or ip range(s)"),
	theServers(this, "servers <str>", "servers ip(s) or ip range(s)"),
	thePeakRate(this, "peak_rate <pkt/sec>", "maximum number of packets sent per second"),
	theStuckTout(this, "stuck_tout <time>", "exit if no new connections established"),
	theNotifAddr(this, "notify <addr>",      "send notifications to a [remote] monitoring program", NetAddr()),
	theConsFileName(this, "console <file>", "redirect console output", "-") {
}

ostream &ConnProbeOpts::printAnonym(ostream &os) const {
	return os << "[clients servers]";
}

bool ConnProbeOpts::parseAnonym(const Array<const char *> &opts) {
	if (!opts.count())
		return true;

	if (opts.count() != 2) {
		cerr << "malformed anonymous options, use --help" << endl;
		return false;
	}

	return 
		theClients.parse(theClients.name(), opts[0]) &&
		theServers.parse(theServers.name(), opts[1]);
}

bool ConnProbeOpts::validate() const {
	if (!theSideName)
		cerr << "no side name specified" << endl;
	else
	if (theSideName != "client" && theSideName != "server")
		cerr << "unknown side name (" << theSideName << ") specified" << endl;
	else
	if (!theClients)
		cerr << "no client addresses specified" << endl;
	else
	if (!theServers)
		cerr << "no server addresses specified" << endl;
	else
		return true;
	return false;
}

static
int in_cksum(unsigned short *ptr1, int size1, unsigned short *ptr2, int size2) {
	register unsigned long sum;
	unsigned short oddbyte;

	sum = 0;
	while (size1 > 0) {
		sum += *ptr1++;
		size1 -= 2;
	}
	Assert(0 == size1);
	while (size2 > 1) {
		sum += *ptr2++;
		size2 -= 2;
	}

	if (size2 == 1) {
		oddbyte = 0;
		*((unsigned char *) &oddbyte) = *(unsigned char *) ptr2;
		sum += oddbyte;
	}
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);
	return ((unsigned short) (~sum));
}

static
int addr2index(const NetAddr &addr) {
	int idx = -1;
	if (!theClients.find(addr, idx)) {
		cerr << here << "unknown client: " << addr << endl;
		exit(-1);
	}
	return idx;
}

static 
int send_tcp(const NetAddr &src, const NetAddr &dst, tcp_seq seq, tcp_seq ack, int flags) {
	struct sockaddr_in sa;
	struct pseudo_hdr phdr;

	struct ip *iphdr = (struct ip *)pktbuf;
	struct tcphdr *tcphdr = (struct tcphdr *)(pktbuf + IPHDRLEN);

	iphdr->ip_vhl = IP_VHL_BORING;
	iphdr->ip_tos = 0xFF;
	iphdr->ip_len = IPHDRLEN + TCPHDRLEN;
	iphdr->ip_id = 0;
	iphdr->ip_off = 0;
	iphdr->ip_ttl = 255;
	iphdr->ip_p = IPPROTO_THIS;
	iphdr->ip_sum = 0;
	iphdr->ip_src = src.addrN();
	iphdr->ip_dst = dst.addrN();

	tcphdr->th_sport = htons(src.port());
	tcphdr->th_dport = htons(dst.port());
	tcphdr->th_seq = htonl(seq);
	tcphdr->th_ack = htonl(ack);
	tcphdr->th_off = TCPHDRLEN>>2;
	tcphdr->th_flags = flags;
	tcphdr->th_win = htons(4096);
	tcphdr->th_sum = 0;
	tcphdr->th_urp = 0;

	phdr.src = src.addrN();
	phdr.dst = dst.addrN();
	phdr.zero = 0;
	phdr.proto = IPPROTO_TCP;
	phdr.len = htons(TCPHDRLEN);

	tcphdr->th_sum = in_cksum(
		(unsigned short *) &phdr, sizeof(phdr),
		(unsigned short *) tcphdr, TCPHDRLEN);
	iphdr->ip_sum = in_cksum(NULL, 0, (unsigned short *) iphdr, IPHDRLEN + TCPHDRLEN);

	memset(&sa, '\0', sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr = dst.addrN();
	npktssent++;
	return sendto(raw.fd(), pktbuf, IPHDRLEN + TCPHDRLEN, 0, (struct sockaddr *) &sa, sizeof(sa));
}

static 
void handle_pkt(int, const NetAddr &) {
	struct ip *iphdr = (struct ip *)pktbuf;
	if (IPPROTO_THIS != iphdr->ip_p)
		return;
	if (0xFF != iphdr->ip_tos)
		return;
	npktsrecv++;
	struct tcphdr *tcphdr = (struct tcphdr *)(pktbuf + IPHDRLEN);
	if (TH_SYN == tcphdr->th_flags) {
		Assert(!isClient);
		/* check for TCPS_CLOSED state? */
		const int index = addr2index(NetAddr(iphdr->ip_src, ntohs(tcphdr->th_sport)));
		ConnState &conn = theConnections[index];
		conn.tcp_state = TCPS_SYN_RECEIVED;
		conn.seq = myseq;
		conn.ack = ntohl(tcphdr->th_seq) + 1;
		const int x = send_tcp(
			NetAddr(iphdr->ip_dst, ntohs(tcphdr->th_dport)),
			NetAddr(iphdr->ip_src, ntohs(tcphdr->th_sport)),
			conn.seq,
			conn.ack,
			TH_SYN|TH_ACK);
		if (Should(x >= 0)) {
			++TheStats.theXactLvl;
			++TheStats.theOpenLvl;
		}
	} else if ((TH_SYN|TH_ACK) == tcphdr->th_flags) {
		Assert(isClient);
		/* check for TCPS_SYN_RECEIVED state? */
		const int index = addr2index(NetAddr(iphdr->ip_dst, ntohs(tcphdr->th_dport)));
		ConnState &conn = theConnections[index];
		conn.tcp_state = TCPS_ESTABLISHED;
		conn.ack = ntohl(tcphdr->th_seq) + 1;
		const int x = send_tcp(
			NetAddr(iphdr->ip_dst, ntohs(tcphdr->th_dport)),
			NetAddr(iphdr->ip_src, ntohs(tcphdr->th_sport)),
			conn.seq,
			conn.ack,
			TH_ACK);
		if (Should(x >= 0)) {
			nestablished++;
			--TheStats.theXactLvl;
			++TheStats.theEstbLvl;
			theLastEstbTime = TheClock;
		}
	} else if ((TH_ACK) == tcphdr->th_flags) {
		Assert(!isClient);
		const int index = addr2index(NetAddr(iphdr->ip_src, ntohs(tcphdr->th_sport)));
		ConnState &conn = theConnections[index];
		conn.tcp_state = TCPS_ESTABLISHED;
		nestablished++;
		--TheStats.theXactLvl;
		++TheStats.theEstbLvl;
		theLastEstbTime = TheClock;
	} else {
		cerr << "unexpected packet" << endl;
		abort();
	}
}

static
int read_pkt() {
	NetAddr from;
	int rl = raw.recvFrom(pktbuf, sizeof(pktbuf), from, 0);
	if (rl < 0) {
		if (EAGAIN == errno)
			return 0;
		Should(false);
		return rl;
	}
	Assert(rl > 0);
	handle_pkt(rl, from);
	return rl;
}

static
int send_pkt(){
	static Time lastSend;
	if (TheConnProbeOpts.thePeakRate.set()) {
		Clock::Update(false);
		if (TheClock.time() - lastSend < TheConnProbeOpts.thePeakRate.intArrTime())
			return 0;
	}

	static AddrHashIter iter = theClients.iterator();
	Assert(!iter.atEnd());

	int res = 0;
	const int index = iter.data();
	ConnState &conn = theConnections[index];
	if (TCPS_ESTABLISHED != conn.tcp_state) {
		const NetAddr &src = iter.addr();
		const NetAddr &dst = *theServers[0];
		conn.seq = myseq;
		conn.ack = 0;
		res = send_tcp(src, dst, conn.seq, conn.ack, TH_SYN);
		if (Should(res >= 0)) {
			++TheStats.theXactLvl;
			++TheStats.theOpenLvl;
		}
		conn.tcp_state = TCPS_SYN_SENT;
		lastSend = TheClock;
	}

	++iter;
	if (iter.atEnd())
		iter.rewind();
	Assert(!iter.atEnd());
	return res;
}

static
void report() {
	static Time lastReport = TheClock;
	static int wasEst = -1;
		
	if (wasEst < 0) {
		clog << "#time     "
			<< ' ' << " pkt_sent "
			<< ' ' << " pkt_recv "
			<< ' ' << "conn_estb "
			<< ' ' << "conn_rate "
			<< endl;
		wasEst = 0;
	}

	if (TheClock.time() > lastReport + Time::Sec(5)) {
		const Time dur = TheStats.theDuration = TheClock-lastReport;
		const double rate = (nestablished - wasEst)/dur.secd();

		clog << TheClock.time().sec()
			<< ' ' << setw(10) << npktssent 
			<< ' ' << setw(10) << npktsrecv
			<< ' ' << setw(10) << nestablished
			<< ' ' << setw(10) << rate
			<< endl;


		if (TheNotifier)
			TheNotifier->noteStats(TheStats, isClient ? lgcCltSide : lgcSrvSide);

		TheStats.restart();
		lastReport = TheClock;
		wasEst = nestablished;
	}

	const Time stuckTime = TheClock.time() - theLastEstbTime;
	if (TheConnProbeOpts.theStuckTout >= 0 && theLastEstbTime >= 0 &&
		stuckTime >= TheConnProbeOpts.theStuckTout) {
		clog << thePrgName << ": fyi: no new connections established for"
			<< " the last " << stuckTime << ", exiting" << endl;
		exit(0);
	}
}

static
void io_loop() {
	for (int count = 0; ; ++count) {
		if ((count++ & 0xFFF) == 0) {
			Clock::Update(false);
			report();
		}
		if (read_pkt() < 0)
			break;
		if (isClient && send_pkt() < 0)
			break;
		myseq += 64000;
	}
}

static
void addCltAddr(const NetAddr &addr) {
	theClients.add(addr, theClients.count());
}

static
bool handleCmdLine(int argc, char *argv[]) {
	Array<OptGrp*> opts;
	opts.append(&TheConnProbeOpts);
	theCmdLine.configure(opts);
	if (!theCmdLine.parse(argc, argv))
		return false;

	// validate command line params
	for (int i = 0; i < opts.count(); ++i) {
		if (!opts[i]->validate())
			return false;
	}

	isClient = TheConnProbeOpts.theSideName == "client";

	/* put command line addresses into theClients hash */

	const Array<String*> &cltAddrs = TheConnProbeOpts.theClients.val();

	// compute the number of addresses first
	int addrCount = 0;
	for (int i = 0; i < cltAddrs.count(); ++i) {
		PglNetAddrRange hostParser;
		// delete port info to get the right count
		const String &host = *cltAddrs[i];
		String addr;
		if (const char *port = host.rchr(':'))
			addr = host(0, port - host.str());
		else
			addr = host;
		if (!hostParser.parse(addr))
			cerr << thePrgName << ": malformed client address: `" << addr << endl << xexit;
		addrCount += hostParser.addrCount();
	}

	theClients.capacity(addrCount + addrCount/3);
	for (int i = 0; i < cltAddrs.count(); ++i) {
		const String &host = *cltAddrs[i];
		PglNetAddrRange hostParser;
		if (!hostParser.parse(host))
			cerr << thePrgName << ": malformed client address: `" << host << endl << xexit;
		hostParser.toAddrs(&addCltAddr);
	}
	Assert(theClients.count() >= addrCount);

	// put command line addresses into theServers array
	const Array<String*> &srvAddrs = TheConnProbeOpts.theServers.val();
	for (int i = 0; i < srvAddrs.count(); ++i) {
		const String &host = *srvAddrs[i];
		PglNetAddrRange hostParser;
		if (!hostParser.parse(host))
			cerr << thePrgName << ": malformed server address: `" << host << endl << xexit;
		hostParser.toAddrs(theServers);
	}
	if (theServers.count() != 1) {
		cerr << thePrgName << ": only one server address can be specified (for now), got "
			<< theServers.count() << endl << xexit;
	}

	if (TheConnProbeOpts.theNotifAddr)
		TheNotifier = new Notifier("connprb", TheConnProbeOpts.theNotifAddr);

	return true;
}

static
void configureLogs(int prec) {
	// redirect console output
	if (TheConnProbeOpts.theConsFileName && strcmp(TheConnProbeOpts.theConsFileName, "-"))
		cout = cerr = clog = *new ofstream(TheConnProbeOpts.theConsFileName);

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}


int main(int argc, char *argv[]) {
	thePrgName = argv[0];
	if (!handleCmdLine(argc, argv))
		return -1;

	configureLogs(2);

	theCmdLine.reportRaw(clog << thePrgName << ": "); clog << endl;
	theCmdLine.reportParsed(clog << thePrgName << ": "); clog << endl;
	clog << thePrgName << ": simulating " 
		<< theClients.count() << " clients and "
		<< theServers.count() << " servers" << endl;
	clog << thePrgName << ": fyi: address hash: max chain: " 
		<< theClients.maxChainLen() << " addresses"
		<< " utilization: " << theClients.utilization() << "%" << endl;

	theConnections.stretch(theClients.count());
	theConnections.count(theConnections.capacity());
	
	Must(raw.create());
	Must(raw.includeIpHdr(true));
	Must(raw.blocking(false));

	RndGen rng(time(0));
	myseq = (tcp_seq) rng.ltrial();

	io_loop();

	raw.close();
	return 0;
}
