
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "xstd/Rnd.h"
#include "xstd/gadgets.h"
#include "base/ConnCloseStat.h"
#include "runtime/ConnIdx.h"
#include "runtime/ConnMgr.h"
#include "runtime/globals.h"



ObjFarm<Connection> ConnMgr::TheConnFarm;
ConnIdx ConnMgr::TheConnIdx;


ConnMgr::ConnMgr(): thePconnUseLmt(0), theConnLvl(0) {
	// init static arrays (could init more than once, but that's ok)
	TheConnFarm.limit(Socket::TheMaxLevel+1);
	TheConnIdx.fdLimit(Socket::TheMaxLevel+1);
}

void ConnMgr::configure(const SockOpt &anOpt, RndDistr *aPconnUseLmt) {
	Assert(!thePconnUseLmt);
	thePconnUseLmt = aPconnUseLmt;
	theSockOpt = anOpt;
}

void ConnMgr::idleTimeout(Time aTout) {
	theIdleTimeout = aTout;
}


void ConnMgr::put(Connection *conn) {
	TheConnIdx.check(conn);

	// see if we want to keep this connection open
	if (conn->reusable())
		putIdle(conn); // may close due to external limits
	else
		closeBusy(conn);
}

// adds freshly opened connection
void ConnMgr::opened(Connection *conn) {
	conn->mgr(this);
	TheConnIdx.add(conn);
	setUseLimit(conn);
	theConnLvl++;
}

void ConnMgr::setUseLimit(Connection *conn) {
	int limit = 1; // no pconns by default
	if (thePconnUseLmt)
		limit = Max(1, (int)thePconnUseLmt->trial());
	conn->useLimit(limit);
}

void ConnMgr::noteTimeout(int fd, Time) {
	Connection *conn = TheConnIdx[fd];
	Assert(conn);
	conn->theRd.stop(this);
	closeIdle(conn, ConnCloseStat::ckIdleLocal);
}

void ConnMgr::putIdle(Connection *conn) {
	// start waiting for a foreign close
	if (theIdleTimeout >= 0)
		conn->theRd.start(this, theIdleTimeout);
	else
		conn->theRd.start(this);
	TheFileScanner->setPriority(conn->sock().fd(), fsupBestEffort);
}

void ConnMgr::closeIdle(Connection *conn, CloseKind ck) {
	conn->closeKind(ck);
	delIdle(conn);
	close(conn);
}

void ConnMgr::closeBusy(Connection *conn) {
	close(conn);
}

// internal method to close a connection, no checks made here
void ConnMgr::close(Connection *conn) {
	TheConnIdx.del(conn);
	conn->close();
	TheConnFarm.put(conn);
	theConnLvl--;
}

